/*
 * irq-rom.c -- BSP interrupt handling needed only for ROM builds.
 *
 * Copyright (c) 1998, 1999 Cygnus Support
 *
 * The authors hereby grant permission to use, copy, modify, distribute,
 * and license this software and its documentation for any purpose, provided
 * that existing copyright notices are retained in all copies and that this
 * notice is included verbatim in any distributions. No written agreement,
 * license, or royalty fee is required for any of the authorized uses.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 */

#include <stdlib.h>
#include <bsp/bsp.h>
#include "../include/debflags.h"
#include "bsp_if.h"

// #ifndef DEBUG_DISPATCH
// #define DEBUG_DISPATCH 1
// #endif


/*
 *  Dispatch function for CPU exceptions.
 *  Returns true if exception was handled by installed handler,
 *  zero if not.
 */
int
_bsp_exc_dispatch(int exc_nr, void *regs)
{
    bsp_vec_t *vec;
    int done;

#if DEBUG_DISPATCH
    bsp_printf("_bsp_exc_dispatch(), exc_nr=%d, regs=%x\n", exc_nr, &vec);
#endif // DEBUG_DISPATCH

    vec = bsp_shared_data->__exc_table[exc_nr];

#if DEBUG_DISPATCH
    bsp_printf("_bsp_exc_dispatch(), vec=%lx, vec->handler=%x\n",
                vec, vec->handler);
#endif // DEBUG_DISPATCH

    done = 0;
    while (!done && vec != NULL && vec->handler != NULL) {
#if DEBUG_DISPATCH
    	bsp_printf("_bsp_exc_dispatch(), handling exception @ 0x%08x\n",
			vec);
#endif // DEBUG_DISPATCH

	done = vec->handler(exc_nr, regs);

#if DEBUG_DISPATCH
    	bsp_printf("_bsp_exc_dispatch(), exception handled, result=%d\n", done);
#endif // DEBUG_DISPATCH
	vec = vec->next;
    }

#if DEBUG_DISPATCH
    if (!done)
    {
        bsp_printf("_bsp_exc_dispatch(), exception NOT handled\n");
    }
#endif // DEBUG_DISPATCH

#ifdef XXX_DEBUG_DISPATCH
    if (!done)
	bsp_printf("Unexpected Exception [%d]\n", exc_nr);
    else
	bsp_printf("_bsp_exc_dispatch(%d) done.\n", exc_nr);
#endif // DEBUG_DISPATCH

    return done;
}


/*
 * Boot-time initialization of bsp interrupt handling module.
 * Should be called from _bsp_late_init after CPU and board
 * specific initialization code has installed interrupt
 * controller descriptors.
 */
void
__init_irq_controllers(void)
{
    int i;
    const struct bsp_irq_controller *ic;

#if DEBUG_IRQ
    bsp_printf("__init_irq_controllers() entry\n");
#endif // DEBUG_IRQ

/*
	write_led (96);
*/
    for (i = 0; i < BSP_MAX_IRQ_CONTROLLERS; i++) {
	ic = bsp_shared_data->__ictrl_table[i];
	if (ic == NULL)
	    break;

#if DEBUG_IRQ
        bsp_printf("__init_irq_controllers(), initializing ic=%lx\n", ic);
#endif // DEBUG_IRQ

	ic->init(ic);
    }
#if DEBUG_IRQ
    bsp_printf("__init_irq_controllers(), done\n");
#endif // DEBUG_IRQ
}


/*
 *  Install an interrupt controller descriptor.
 */
void
_bsp_install_irq_controller(const struct bsp_irq_controller *ic)
{
    int i;

#if DEBUG_IRQ
    bsp_printf("_bsp_install_irq_controller() entry\n");
#endif // DEBUG_IRQ

    for (i = 0; i < BSP_MAX_IRQ_CONTROLLERS; i++) {
	if (bsp_shared_data->__ictrl_table[i] == NULL)
	    break;
    }

    if (i < BSP_MAX_IRQ_CONTROLLERS)
    {
	bsp_shared_data->__ictrl_table[i] = ic;

#if DEBUG_IRQ
    bsp_printf("_bsp_install_irq_controller() __ictrl_table[%d] = %lx\n",
        i, ic);
#endif // DEBUG_IRQ
    }
}

