/*
 * ip.c -- Polling IP for Cygnus BSP
 *
 * Copyright (c) 1998, 1999 Cygnus Support
 *
 * The authors hereby grant permission to use, copy, modify, distribute,
 * and license this software and its documentation for any purpose, provided
 * that existing copyright notices are retained in all copies and that this
 * notice is included verbatim in any distributions. No written agreement,
 * license, or royalty fee is required for any of the authorized uses.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 */
#include <bsp/bsp.h>
#include "bsp_if.h"
#include "net.h"

/* #define TEST_ONLY_GET_OUR_IP_PACKET	1  */
// #define DEBUG_IP_SEND
// #define DEBUG_IP_HANDLER
//#define BSP_PRINTOUT_UDP_INFO 1

ip_addr_t __local_ip_addr;

static word ip_ident;

#if IP_STATS
#endif

void __tcp_handler(pktbuf_t *pkt, ip_route_t *r);
void __udp_handler(pktbuf_t *pkt, ip_route_t *r);
void __icmp_handler(pktbuf_t *pkt, ip_route_t *r);

/*
 * Match given IP address to our address.
 * Check for broadcast matches as well.
 */
static int
ip_addr_match(ip_addr_t addr)
{
    if (addr[0] == 255 && addr[1] == 255 && addr[2] == 255 && addr[3] == 255)
	return 1;

    if (!memcmp(addr, __local_ip_addr, sizeof(ip_addr_t)))
	return 1;

    /*
     * Consider it an address match if we haven't gotten our IP address yet.
     * Some DHCP servers will address IP packets to the assigned address
     * instead of a IP broadcast address.
     */
    if (__local_ip_addr[0] == 0 && __local_ip_addr[1] == 0 &&
	__local_ip_addr[2] == 0 && __local_ip_addr[3] == 0)
	return 1;

    return 0;
}


/*
 * Handle IP packets coming from the polled ethernet interface.
 */
void
__ip_handler(pktbuf_t *pkt, enet_addr_t src_enet_addr)
{
    ip_header_t *ip = pkt->ip_hdr;
    ip_route_t  r;
    int         hdr_bytes;

#ifdef BSP_PRINTOUT_DEBUG_INFO
    bsp_printf(" in iph ip_hdr_len=%d, \n", ip->hdr_len);
#endif /* #ifdef BSP_PRINTOUT_DEBUG_INFO */
#ifdef BSP_PRINTOUT_UDP_INFO 1
    bsp_printf(" in iph ip_hdr_len=%d, \n", ip->hdr_len);
#endif //#ifdef BSP_PRINTOUT_UDP_INFO 1

        /* first make sure its ours and has a good checksum. */
    if (!ip_addr_match(ip->destination) ||
	__sum((word *)ip, ip->hdr_len << 2, 0) != 0) {
	__pktbuf_free(pkt);
        return;
    }

    memcpy(r.ip_addr, ip->source, sizeof(ip_addr_t));
    memcpy(r.enet_addr, src_enet_addr, sizeof(enet_addr_t));

    hdr_bytes = ip->hdr_len << 2;
    pkt->pkt_bytes = ntohs(ip->length) - hdr_bytes;

    switch (ip->protocol) {

#if NET_SUPPORT_ICMP
      case IP_PROTO_ICMP:

#ifdef BSP_PRINTOUT_DEBUG_INFO
     bsp_printf(" in iphdler ICMP rcvd\n"); /* DC temp debug */
#endif /* #ifdef BSP_PRINTOUT_DEBUG_INFO */

	pkt->icmp_hdr = (icmp_header_t *)(((char *)ip) + hdr_bytes);
	__icmp_handler(pkt, &r);
	break;
#endif


#if NET_SUPPORT_TCP
      case IP_PROTO_TCP:

#ifdef BSP_PRINTOUT_DEBUG_INFO
     bsp_printf(" in iphdler TCP rcvd\n"); /* DC temp debug */
#endif /* #ifdef BSP_PRINTOUT_DEBUG_INFO */

	pkt->tcp_hdr = (tcp_header_t *)(((char *)ip) + hdr_bytes);
	__tcp_handler(pkt, &r);
	break;
#endif

#if NET_SUPPORT_UDP
      case IP_PROTO_UDP:

#ifdef BSP_PRINTOUT_UDP_INFO
     bsp_printf(" in iphdler UDP rcvd, call __udp_handler\n"); //DC temp debug
#endif //#ifdef BSP_PRINTOUT_UDP_INFO

	pkt->udp_hdr = (udp_header_t *)(((char *)ip) + hdr_bytes);
	__udp_handler(pkt, &r);
	break;
#endif

      default:
	__pktbuf_free(pkt);
	break;
    }
}


/*
 * Send an IP packet.
 *
 * The IP data field should contain pkt->pkt_bytes of data.
 * pkt->[udp|tcp|icmp]_hdr points to the IP data field. Any
 * IP options are assumed to be already in place in the IP
 * options field.
 */
void
__ip_send(pktbuf_t *pkt, int protocol, ip_route_t *dest)
{
    ip_header_t *ip = pkt->ip_hdr;
    int         hdr_bytes;

    /*
     * Figure out header length. The use udp_hdr is
     * somewhat arbitrary, but works because it is
     * a union with other IP protocol headers.
     */
    hdr_bytes = ((char *)pkt->udp_hdr) - ((char *)ip);

    pkt->pkt_bytes += hdr_bytes;

#ifdef DEBUG_IP_SEND
	bsp_printf("ip_send pkt->pkt_bytes %d\n",pkt->pkt_bytes);
#endif // DEBUG_IP_SEND

    ip->version = 4;
    ip->hdr_len = hdr_bytes >> 2;

#ifdef DEBUG_IP_SEND
    bsp_printf("ip_send ip->hdr_len %d\n",ip->hdr_len);
#endif // DEBUG_IP_SEND

    ip->tos = 0;
    ip->length = htons(pkt->pkt_bytes);
    ip->ident = ip_ident++;
    ip->fragment = 0;
    ip->ttl = 255;
    ip->protocol = protocol;
    ip->checksum = 0;

#ifdef BSP_PRINTOUT_DEBUG_INFO
    bsp_printf("ip_send: local_ip %x %x %x %x\n",
        (unsigned) __local_ip_addr[0],
        (unsigned) __local_ip_addr[1],
        (unsigned) __local_ip_addr[2],
        (unsigned) __local_ip_addr[3]);
#endif /* #ifdef BSP_PRINTOUT_DEBUG_INFO */

    memcpy(ip->source, __local_ip_addr, sizeof(ip_addr_t));
    memcpy(ip->destination, dest->ip_addr, sizeof(ip_addr_t));
/* DC need to convert to make it right 
    ip->checksum = __sum((word *)ip, hdr_bytes, 0); */
    ip->checksum = ntohs(__sum((word *)ip, hdr_bytes, 0));
#ifdef BSP_PRINTOUT_DEBUG_INFO
     bsp_printf(" ip_send req %d bytes\n", htons(ip->length) );
#endif /* #ifdef BSP_PRINTOUT_DEBUG_INFO */
/* D temp */ 

#ifdef USE_HARDWIRED_MAC_ADDR
    dest->enet_addr[0]=0x00;
    dest->enet_addr[1]=0xD0;
    dest->enet_addr[2]=0xB7;
    dest->enet_addr[3]=0x2E;
    dest->enet_addr[4]=0x0F;
    dest->enet_addr[5]=0xA4;
#endif // USE_HARDWIRED_MAC_ADDR

    __enet_send(pkt, dest->enet_addr, ETH_TYPE_IP);    

#ifdef BSP_PRINTOUT_DEBUG_INFO
    bsp_printf("__ip_send returning\n");
#endif /* #ifdef BSP_PRINTOUT_DEBUG_INFO */
}


