/*
 * breakpoint.c
 *
 * Copyright (c) 1998, 1999 Cygnus Solutions
 *
 * The authors hereby grant permission to use, copy, modify, distribute,
 * and license this software and its documentation for any purpose, provided
 * that existing copyright notices are retained in all copies and that this
 * notice is included verbatim in any distributions. No written agreement,
 * license, or royalty fee is required for any of the authorized uses.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 */
#include <stdlib.h>
#ifdef HAVE_BSP
#include <bsp/bsp.h>
#include <bsp/cpu.h>
#endif

#include "monitor.h"
#include "tservice.h"
#include "stub-tservice.h"

#include "fmt_util.h"


static struct bp *last_bp_ptr;
static struct bp *first_bp_ptr;

#ifdef NO_MALLOC
static struct bp *free_bp_list;
static struct bp bp_list[MAX_NUM_BP];
static int       curr_bp_num;
#endif

int
add_mon_breakpoint (mem_addr_t location)
{
  struct bp *ptr;
  struct bp *new_bp_ptr;

  for (ptr = first_bp_ptr; ptr != NULL; ptr = ptr->next)
    {
      if (MEM_ADDR_EQ_P (ptr->address, location))
	return 1;
    }
#ifdef NO_MALLOC
  if (free_bp_list != NULL)
    {
      new_bp_ptr = free_bp_list;
      free_bp_list = new_bp_ptr->next;
    }
  else
    {
      if (curr_bp_num < MAX_NUM_BP)
	{
	  new_bp_ptr = &bp_list[curr_bp_num++];
	}
      else
	{
	  xprintf ("No more breakpoints\n");
	  return 1;
	}
    }
#else
  new_bp_ptr = (struct bp *)malloc (sizeof (struct bp));
#endif

  if (first_bp_ptr == NULL)
    {
      first_bp_ptr = new_bp_ptr;
    }
  else
    {
      last_bp_ptr->next = new_bp_ptr;
    }
  last_bp_ptr = new_bp_ptr;

  last_bp_ptr->next = NULL;
  last_bp_ptr->address = location;
  last_bp_ptr->in_memory = 0;
  return 0;
}


void
install_breakpoints (void)
{
  struct bp *ptr = first_bp_ptr;
  while (ptr != NULL)
    {
      set_breakpoint (ptr);
      ptr = ptr->next;
    }
}


void
clear_breakpoints (void)
{
  struct bp *ptr = first_bp_ptr;

  while (ptr != NULL)
    {
      clear_breakpoint (ptr);
      ptr = ptr->next;
    }
}

int
show_breakpoints (void)
{
  struct bp *ptr;

  for (ptr = first_bp_ptr; ptr != NULL; ptr = ptr->next)
    {
      char buf[20];

      addr2str (&ptr->address, buf);
      xprintf ("%s\n", buf);
    }
  
  return 0;
}



int 
clear_mon_breakpoint (mem_addr_t location)
{
  int error = 0;
  struct bp *ptr = first_bp_ptr;
  struct bp *prev_ptr = NULL;

  /* Scan the list looking for the address to clear */
  while (ptr != NULL && !MEM_ADDR_EQ_P (ptr->address, location))
    {
      /* keep a pointer one behind the current position */
      prev_ptr = ptr;
      ptr = ptr->next;
    }
  if (ptr == NULL)
    {
      xprintf ("That address has no breakpoint on it.\n");
      error = 1;
    }
  else
    {
      /* Just in case it's still in memory. */
      clear_breakpoint (ptr);

      /* now we'll point the previous bp->next at the one after the one 
	 we're deleting, unless there is no previous bp. */
      if (prev_ptr != NULL)
	{
	  prev_ptr->next = ptr->next;
	}

      if (first_bp_ptr == ptr)
	first_bp_ptr = ptr->next;

      if (last_bp_ptr == ptr)
	last_bp_ptr = prev_ptr;

      /* eliminate the offending bp struct */
#ifdef NO_MALLOC
      ptr->next = free_bp_list;
      free_bp_list = ptr;
#else
      free (ptr);
#endif
    }
  return error;
}	

